/* A simple program to calculate worst-case response times for tasks
 * with arbitrary deadlines and offsets, including tick overheads
 * 
 * Copyright (C) 1992 Ken Tindell, 
 * Department of Computer Science, University of York,
 * e-mail: ken@minster.york.ac.uk
 *
 * program reads in a file with the following format:
 *
 * parameters <Tclk> <Cclk> <CQL> <CQS> <max resp>
 * trans <transID> Ti
 * ...
 * task <ID> <transID> <ei> <Di> <Ci> <Ji> <Oi>
 * ...
 * sem <ID>
 * ...
 * lock <semID> by <taskID> <wcet>
 * ...
 * precedence <taskID> <taskID>
 * exclusion <taskID> <taskID>
 *
 * the file lists tasks in priority order by default. the -O switch
 * invokes the priority optimiser. the -E switch makes the test used exact.
 *
 * The flag -pnnnnn indicates the output parameters that are desired, from
 * the following:

 * r, response time
 * R, original analysis response time
 * D, deadline
 * T, transaction period
 * e, e value
 * B, blocking factor
 * C, WCET
 * O, offset
 * J, release jitter
 * s, semaphore information
 * o, object information
 * t, task/transaction information
 * c, constraint satisfaction information
 * 
 * For example, -prDeTC will output the response time, deadline, e, T, and C
 * for printed tasks/transactions, in that order.
 *
 * The flag -W <wordfile> will output the data separated by commas in the
 * specified file. In this format the data can be cut&pasted into Word
 * tables.
 * The flag -o <outputfile> will output the data to the specified file;
 * default is to write to standard output
 * The flag -i <inputfile> will take data from the specified file
 * The flag -E will apply the exact offset test (can take a long time!)
 * The flag -C applies the scheduling overheads model
 * The flag -w sets a window size over which to measure scheduling overheads
 * The flag -S sets the time the simulation of the system should run for
 * (the default is zero, i.e. don't simulate)
 * The flag -I calculates the idle task response time (i.e. the longest
 * busy period)
 *
 */

#include "offsets.h"

excl_attributes excl_map[MAX_EXCLS];
prec_attributes prec_map[MAX_PRECS];
trans_attributes trans_map[MAX_TRANS];
task_attributes task_map[MAX_TASKS];
sem_attributes sem_map[MAX_SEMS];
lock_attributes lock_map[MAX_LOCKS];
message_attributes message_map[MAX_MESSAGES];
cpu_attributes cpu_map[MAX_CPUS];
net_attributes net_map[MAX_NETS];

int num_trans;
int num_tasks;
int num_sems;
int num_locks;
int num_precs;
int num_excls;
int num_messages;
int num_cpus;
int num_nets;
timeval max_resp;					  /* maximum response time - cut the window off at this time */
int idle_trans;					  /* the idle transaction, containing ... */
int idle_task;						  /* ... the idle task */
timeval clock_tick;

timeval Cclk = 0;
timeval CQL = 0;
timeval CQS = 0;
timeval Tclk = 1000;
timeval tick_window = 10000;
timeval sim_ticks = 10000;

double utilisation = 0.0;

int add_tick_overheads_flag = FALSE;
int calculate_idle_task = FALSE;
int tractable_test = TRUE;
static char default_output[] = "rRDCeTOJBtc";

/* forward declarations */

static void findmaxW();
static timeval findr();
timeval find_arrival_time();
static timeval findw();
static timeval wnplus();
static timeval I();
timeval simpledms();
timeval tick_overheads();
void print_task_attribute();
void print_message_attribute();
int load_tasks();
void set_all_Bi();
void set_net_Bi();
int check_prec();
int check_excl();
void simulate();
void print_details();
int ceildiv();
int max();
static void firstWvalue();
static int nextWvalue();
static int permuteWvalue();
static timeval alltransI();
static int K();
static int v();
timeval simple_tick_overheads();
int floordiv();
int find_task();
int find_cpu();
int find_trans();
int find_sem();
int find_net();
void sim_init();
void sim_stretch();

extern int getopt();
extern char *optarg;
extern int optind;
extern int opterr;

FILE *output_file;
FILE *input_file;
FILE *word_file;

char *output_reqd;

int main(argc, argv)
int argc;
char *argv[];
{
	int i;
	int keep_iterating;
	int ch;
	char *output_file_name;
	char *input_file_name;
	char *word_file_name;

	output_reqd = default_output;

	output_file_name = NULL;
	input_file_name = NULL;
	word_file_name = NULL;

	while ((ch = getopt(argc, argv, "ECIOo:i:p:W:w:S:")) != EOF) {
		switch (ch) {
		case 'E':
			tractable_test = FALSE;
			break;
		case 'S':
			sim_ticks = atoi(optarg);
			break;
		case 'w':
			tick_window = atoi(optarg);
			break;
		case 'C':
			add_tick_overheads_flag = TRUE;
			break;
		case 'o':
			output_file_name = optarg;
			break;
		case 'I':
			calculate_idle_task = TRUE;
			break;
		case 'i':
			input_file_name = optarg;
			break;
		case 'p':
			output_reqd = optarg;
			break;
		case 'W':
			word_file_name = optarg;
			break;
		}
	}

	if (input_file_name) {
		input_file = fopen(input_file_name, "r");
		if (input_file == NULL) {
			fprintf(stderr,
					  "error trying to open %s for reading\n", input_file_name);
			exit(1);
		}
	} else {
		input_file = NULL;
	}
	if (output_file_name) {
		output_file = fopen(output_file_name, "w");
		if (output_file == NULL) {
			fprintf(stderr,
					  "error trying to open %s for writing\n",
					  output_file_name);
			exit(1);
		}
	} else {
		output_file = stdout;
	}

	if (word_file_name) {
		word_file = fopen(word_file_name, "w");
		if (output_file == NULL) {
			fprintf(stderr,
					  "error trying to open %s for writing\n", word_file_name);
			exit(1);
		}
	}

	printf("worst-case response time calculation..\n");

	if (input_file != NULL) {
		if (load_tasks(input_file)) {
			fprintf(stderr, "Errors in the input file\n");
			exit(1);
		}

		set_all_Bi();
		set_net_Bi();

		do {
			keep_iterating = FALSE;
			for (i = 0; i < num_tasks - 1; i++) {
				printf("finding task %i response time\n", i);
				if (task_map[i].receives != -1) {
					task_map[i].O =
						 message_map[task_map[i].receives].O +
						 message_map[task_map[i].receives].r;
				}
				findr(i);
				task_map[i].dmsr = simpledms(i);
				if (task_map[i].dmsr < task_map[i].r) {
					fprintf(stderr,
							  "assertion failure: new test %d, old test %d\n",
							  task_map[i].r, task_map[i].dmsr);
				}

				if (task_map[i].r > task_map[i].oldr) {
					keep_iterating = TRUE;
				}
				if (task_map[i].r < task_map[i].oldr) {
					fprintf(stderr, "outer iteration not monotonic (tasks)!\n");
					keep_iterating = TRUE;
					/* abort(); */
				}
				task_map[i].oldr = task_map[i].r;
			}

			for (i = 0; i < num_messages; i++) {
				message_map[i].O =
					 task_map[message_map[i].source].O +
					 task_map[message_map[i].source].r;
				message_map[i].r = find_arrival_time(i);
				if (message_map[i].r > message_map[i].oldr) {
					keep_iterating = TRUE;
				}
				if (message_map[i].r < message_map[i].oldr) {
					fprintf(stderr, "outer iteration not monotonic (messages)!\n");
					keep_iterating = TRUE;
					/* abort; */
				}
				message_map[i].oldr = message_map[i].r;
			}
			for (i = 0; i < num_tasks - 1; i++) {
				if (task_map[i].receives != -1) {
					if (task_map[i].O !=
						 message_map[task_map[i].receives].O +
						 message_map[task_map[i].receives].r)
					keep_iterating = TRUE;
				}
			}
			for (i = 0; i < num_messages; i++) {
				if (message_map[i].O !=
					 task_map[message_map[i].source].O +
					 task_map[message_map[i].source].r)
				keep_iterating = TRUE;
			}
		} while (keep_iterating);

		if (calculate_idle_task) {
			findr(num_tasks - 1);
		}

		for (i = 0; i < num_precs; i++) {
			prec_map[i].result = check_prec(prec_map[i].a, prec_map[i].b);
		}
		for (i = 0; i < num_excls; i++) {
			excl_map[i].result = check_excl(excl_map[i].a, excl_map[i].b);
		}

		simulate(sim_ticks);		  /* duration of simulation */

		print_details(output_file);
	}
	return 1;
}

static int find_corresp_r(i, t, q)
int i, t, q;
{
	int kj;
	timeval Oi, Ji, Wt, Tt;
	int ei;
	timeval tmp_ri;
	timeval w;

	Oi = task_map[i].O;
	Ji = task_map[i].J;
	ei = task_map[i].e;
	Tt = trans_map[t].T;

	Wt = trans_map[t].W;
	kj = ceildiv(Wt - Oi - Ji, Tt);
	w = findw(i, q);
	tmp_ri = w + Wt - Tt * (q * ei + kj) - Oi;
	task_map[i].r = max(task_map[i].r, tmp_ri);

	/* if the busy period finishes after the latest release of i
	 * then we need to look further
	 */
	if (w + Wt > Tt * ((q + 1) * ei + kj) + Oi) {
		return 1;
	} else {
		return 0;
	}
}

/* find worst-case response time of task i */
static timeval findr(i)
int i;
{
	int q;							  /* defined by the analysis */
	int t;							  /* the transaction of which task i is a member */
	int must_continue;			  /* set to false when termination condition met */

	/* see technical note: all we need to do here is find w over significant
	 * busy periods
	 */
	task_map[i].r = 0;
	t = task_map[i].trans;
	q = 0;
	do {
		int x;

		must_continue = FALSE;
		if (tractable_test) {
			firstWvalue(t);
			do {
				must_continue = find_corresp_r(i, t, q);
			}
			while (nextWvalue(t));
		} else {
			for (x = 0; x < num_trans; x++) {
				firstWvalue(x);
			}
			/* iterate over all tasks in transaction t */
			do {
				must_continue = find_corresp_r(i, t, q);
			}
			while (permuteWvalue(0));
		}
		q++;
	}
	while (must_continue);

	/* ri cannot be less than Ci */
	if (task_map[i].r < task_map[i].C) {
		fprintf(stderr,
				  "assertion failure: r(%d) is < Ci(%d) (%d)\n",
				  i, task_map[i].C, task_map[i].r);
	}

	return task_map[i].r;
}

/* iterates to find a value for w; will return w=max_resp if w exceeds
 * max_resp (this is needed to cut short overly long iterations when the
 * utilisation is close to 100%)
 *
 * i is the task in "level i busy period", q is defined by the analysis
 * and Wt is defined by the analysis
 */
static timeval findw(i, q)
int i, q;
{
	timeval w, wnp;
	timeval Ci = task_map[i].C;

	w = Ci;
	wnp = wnplus(i, q, w);
	while (w != wnp && w < max_resp) {
		w = wnp;
		wnp = wnplus(i, q, w);
	}

	if (w > max_resp) {
		w = max_resp;
	}

	return w;
}


/* return w(n+1): see technical note for details
 * takes variable i, q value, and Wt value
 */
static timeval wnplus(i, q, w)
int i, q;
timeval w;
{
	timeval Ci, Bi;
	timeval sum;
	int t;

	Ci = task_map[i].C;
	Bi = task_map[i].B;

	t = task_map[i].trans;

	sum = Ci + q * Ci + Bi;

	/* sum up over this transaction */
	sum += I(t, i, w);

	/* sum up over the rest of the transactions */
	sum += alltransI(t, i, w);
	/* add in clock overheads corresponding to set W values */
	sum += tick_overheads(w);

	return sum;
}

/* find interference from all transactions other than t over level
 * i busy period of width w
 */
static timeval alltransI(t, i, w)
timeval w;
int i, t;
{
	timeval sum = 0;
	int x;

	for (x = 0; x < num_trans; x++) {
		if (x != t) {
			/* only need to find the value of W if using tractable test */
			if (tractable_test) {
				findmaxW(x, i, w);  /* set Wx to give largest interference */
			}
			sum += I(x, i, w);
		}
	}
	return sum;
}

/* set W of transaction t to first value */
static void firstWvalue(t)
{
	int k;

	if (trans_map[t].num_tasks == 0) {
		fprintf(stderr,
				  "error: transaction %s has no member tasks!\n",
				  trans_map[t].ID);
		return;
	}
	trans_map[t].task_num = 0;
	k = trans_map[t].task_map[0];
	trans_map[t].W = task_map[k].O + task_map[k].J;
}

/* set W of transaction t to the next value, return 1 if more, 0 if reached end
*/
static int nextWvalue(t)
{
	int k;

	trans_map[t].task_num++;
	if (trans_map[t].task_num == trans_map[t].num_tasks) {
		return 0;
	}
	k = trans_map[t].task_map[trans_map[t].task_num];
	trans_map[t].W = task_map[k].O + task_map[k].J;

	return 1;
}

/* find another W permutation for transaction x, not including transaction
 * t; return 1 if still more permutations, 0 if no more.
 */
static int permuteWvalue(x)
int x;
{
	if (!nextWvalue(x)) {		  /* wrap around */
		firstWvalue(x);
		x++;
		if (x >= num_trans) {
			return 0;				  /* finished the lot */
		} else {
			return permuteWvalue(x);
		}
	} else {
		return 1;
	}
}

/* set the value of W for transaction t that gives maximal interference 
 * over w for tasks of higher priority than task i, including
 * tick overheads
 */
static void findmaxW(t, i, w)
int t, i;
timeval w;
{
	timeval maxW;
	timeval maxI;
	int x;

	maxI = 0;
	maxW = 0;
	for (x = 0; x < trans_map[t].num_tasks; x++) {
		int j;
		int Oj, Jj;
		int curI;

		j = trans_map[t].task_map[x];

		Oj = task_map[j].O;
		Jj = task_map[j].J;

		trans_map[t].W = Oj + Jj;
		curI = I(t, i, w) + tick_overheads(w);
		if (curI > maxI) {
			maxI = curI;
			maxW = trans_map[t].W;
		}
	}
	trans_map[t].W = maxW;
}

/* return the total interference from transaction t on a level i busy
 * period of width w
 */
static timeval I(t, i, w)
int t, i;
timeval w;
{
	int sum;
	int x;

	sum = 0;
	for (x = 0; x < trans_map[t].num_tasks; x++) {
		int j;

		j = trans_map[t].task_map[x];

		if (task_map[j].priority < task_map[i].priority
			 && task_map[j].host == task_map[i].host) {
			int Cj;

			Cj = task_map[j].C;
			sum += K(j, t, w) * Cj;
		}
	}

	return sum;
}

static int K(j, t, w)
			/* defined by the analysis; j is a member of transaction t */
int j, t;
timeval w;
{
	timeval Tt, Wt, Oj;
	int ej;

	Tt = trans_map[t].T;
	Wt = trans_map[t].W;
	Oj = task_map[j].O;
	ej = task_map[j].e;

	return ceildiv(Wt + w - Oj - v(j, t) * Tt, ej * Tt);
}

/* define by analysis; j is a member of transaction t */
static int v(j, t)
int j, t;
{
	timeval Wt, Tt, Oj, Jj;

	Wt = trans_map[t].W;
	Tt = trans_map[t].T;
	Oj = task_map[j].O;
	Jj = task_map[j].J;

	return ceildiv(Wt - Oj - Jj, Tt);
}

/* find the worst-case response time of task i, using the simple original
 * analysis
 */
timeval simpledms(i)
{
	int j;
	int q;
	timeval wn, wnp;
	timeval Ci, Bi, Ji, Ti;
	int r = 0;

	Ci = task_map[i].C;
	Bi = task_map[i].B;
	Ji = task_map[i].J;
	Ti = task_map[i].T;
	q = 0;
	do {
		wnp = Ci;
		do {
			wn = wnp;
			wnp = (q + 1) * Ci + Bi + simple_tick_overheads(wn);
			for (j = 0; j < num_tasks; j++) {
				if (task_map[j].priority <
					 task_map[i].priority
					 && task_map[j].host == task_map[i].host) {
					timeval Cj, Jj, Tj;

					Cj = task_map[j].C;
					Jj = task_map[j].J;
					Tj = task_map[j].T;

					wnp += ceildiv(Jj + wn, Tj) * Cj;
				}
			}
		} while (wn != wnp && wn < max_resp);
		r = Ji + max(r, wn - q * Ti);
		q++;
	}
	while (wn > q * Ti - Ji);

	if (r > max_resp) {
		r = max_resp;
	}

	return r;
}

/* ceil of a/b */
int ceildiv(a, b)
int a, b;
{
	if (b <= 0) {
		fprintf(stderr, "ceildiv: not defined on denominator\n");
		exit(1);
	}

	if (a < 0) {
		return -floordiv(-a, b);
	}

	if (a % b == 0) {
		return a / b;
	} else {
		return (a / b) + 1;
	}
}

int floordiv(a, b)
int a, b;
{
	return a / b;
}

/* load the task_map table from the file (from stdin) */
int load_tasks(stream)
FILE *stream;
{
	int errors = 0;
	int Oi;
	int Ti;
	int Di;
	int Ci;
	int Ji;
	int ei;
	char ID[MAX_STRING];
	char keyword[MAX_STRING];
	int line_number;
	char line[MAX_LINE];

	num_trans = 0;
	num_tasks = 0;
	num_sems = 0;
	num_locks = 0;
	num_precs = 0;
	num_excls = 0;
	num_cpus = 0;
	num_nets = 0;
	line_number = 0;
	while (fgets(line, MAX_LINE, stream)) {
		line_number++;
		if (line[0] == '\0') {
			keyword[0] = '\0';
		} else {
			sscanf(line, "%s", keyword);
		}
		if (!strcmp("processor", keyword)) {
			sscanf(line, "%s %s", keyword, ID);

			strcpy(cpu_map[num_cpus].ID, ID);
			num_cpus++;
		}
		if (!strcmp("network", keyword)) {
			char net_type[MAX_STRING];	/* Unused, for now */
			sscanf(line, "%s %s %s", keyword, ID, net_type);

			strcpy(net_map[num_nets].ID, ID);
			strcpy(net_map[num_nets].type, net_type);
			num_nets++;
		}
		if (!strcmp("message", keyword)) {
			char source[MAX_STRING];
			char dest[MAX_STRING];
			char net[MAX_STRING];
			char trans_ID[MAX_STRING];
			int num_dests;
			int the_dest_task, the_source_task, the_trans, the_net;
			int index, index2;

			sscanf(line, "%s %s %s %s %d %d %d %d %s %n",
					 keyword, ID, trans_ID, net, &ei, &Di, &Ci,
					 &Ji, source, &index);

			message_map[num_messages].e = ei;
			message_map[num_messages].D = Di;
			message_map[num_messages].C = Ci;
			message_map[num_messages].J = Ji;
			message_map[num_messages].O = Oi;
			message_map[num_messages].priority = num_messages;
			message_map[num_messages].oldr = 0;

			strcpy(message_map[num_messages].ID, ID);

			the_net = find_net(net);
			if (the_net < 0) {
				fprintf(stderr,
						  "unknown network `%s' line %d\n", net, line_number);
				errors = TRUE;
				goto Errors;
			}

			/* hook up the task to the network */
			message_map[num_messages].net = the_net;

			the_source_task = find_task(source);
			if (the_source_task < 0) {
				fprintf(stderr,
						  "unknown task `%s' line %d\n", source, line_number);
				errors = TRUE;
				goto Errors;
			}
			message_map[num_messages].source = the_source_task;

			num_dests = 0;
			do {
				dest[0] = '\0';
				sscanf(line + index, "%s %n", dest, &index2);
				index += index2;
				if (dest[0] != '\0') {
					the_dest_task = find_task(dest);
					if (the_dest_task < 0) {
						fprintf(stderr,
								  "unknown task %s' line %d\n", dest, line_number);
						errors = TRUE;
						goto Errors;
					}
					if (task_map[the_dest_task].host ==
						 task_map[the_source_task].host) {
						fprintf(stderr,
								  "warning: message `%s' not remote, line %d\n",
								  ID, line_number);
					}
					if (task_map[the_dest_task].receives != -1) {
						fprintf(stderr,
								  "481: can't receive multiple messages, line %d\n",
								  line_number);
						errors = TRUE;
						goto Errors;
					}

					task_map[the_dest_task].receives = num_messages;
					message_map[num_messages].dest[num_dests] = the_dest_task;
					num_dests++;
				}
			} while (dest[0] != '\0');
			message_map[num_messages].num_dests = num_dests;

			the_trans = find_trans(trans_ID);
			if (the_trans == -1) {
				fprintf(stderr,
						  "error: unknown transaction ID line %d\n",
						  line_number);
			} else {
				/* if Oi > Ti then an error */
				Ti = trans_map[the_trans].T;
				if (Oi > Ti) {
					fprintf(stderr,
							  "error: Oi(=%d)>Ti(=%d), message %s, line %d\n",
							  Oi, Ti, ID, line_number);
					errors = TRUE;
				}
				/* need to find transaction period so that task period can
				 * be set; note that task period is equal to ei*Tt
				 */
				message_map[num_messages].T = Ti * ei;
			}
			message_map[num_messages].trans = the_trans;
			trans_map[the_trans].message_map[trans_map[the_trans].num_messages] = num_messages;
			trans_map[the_trans].num_messages++;

			num_messages++;
		}
		if (!strcmp("exclusion", keyword)) {
			char task_A_ID[MAX_STRING];
			char task_B_ID[MAX_STRING];
			int task_a, task_b;

			sscanf(line, "%s %s %s", keyword, task_A_ID, task_B_ID);
			task_a = find_task(task_A_ID);
			if (task_a < 0) {
				fprintf(stderr,
						  "unknown task `%s' line %d\n", task_A_ID, line_number);
				errors = TRUE;
				goto Errors;
			}

			task_b = find_task(task_B_ID);
			if (task_b < 0) {
				fprintf(stderr,
						  "unknown task `%s' line %d\n", task_B_ID, line_number);
				errors = TRUE;
				goto Errors;
			}

			if (task_map[task_a].trans != task_map[task_b].trans) {
				fprintf(stderr,
						  "tasks %s and %s not in the same trans, line %d!\n",
						  task_map[task_a].ID, task_map[task_b].ID, line_number);
				errors = TRUE;
				goto Errors;
			}

			excl_map[num_excls].a = task_a;
			excl_map[num_excls].b = task_b;
			num_excls++;
		}
		if (!strcmp("precedence", keyword)) {
			char task_A_ID[MAX_STRING];
			char task_B_ID[MAX_STRING];
			int task_a, task_b;

			sscanf(line, "%s %s %s", keyword, task_A_ID, task_B_ID);
			task_a = find_task(task_A_ID);
			if (task_a < 0) {
				fprintf(stderr,
						  "unknown task `%s' line %d\n", task_A_ID, line_number);
				errors = TRUE;
				goto Errors;
			}

			task_b = find_task(task_B_ID);
			if (task_b < 0) {
				fprintf(stderr,
						  "unknown task `%s' line %d\n", task_B_ID, line_number);
				errors = TRUE;
				goto Errors;
			}

			if (task_map[task_a].trans != task_map[task_b].trans) {
				fprintf(stderr,
						  "tasks %s and %s not in the same trans, line %d!\n",
						  task_map[task_a].ID, task_map[task_b].ID, line_number);
				errors = TRUE;
				goto Errors;
			}

			prec_map[num_precs].a = task_a;
			prec_map[num_precs].b = task_b;
			num_precs++;
		}
		if (!strcmp("task", keyword)) {
			char host[MAX_STRING];
			char trans_ID[MAX_STRING];
			int the_trans;
			int the_host;

			sscanf(line, "%s %s %s %s %d %d %d %d %d",
					 keyword, ID, trans_ID, host, &ei, &Di, &Ci, &Ji, &Oi);
			task_map[num_tasks].e = ei;
			task_map[num_tasks].D = Di;
			task_map[num_tasks].C = Ci;
			task_map[num_tasks].J = Ji;
			task_map[num_tasks].O = Oi;
			task_map[num_tasks].priority = num_tasks;
			task_map[num_tasks].oldr = 0;
			task_map[num_tasks].receives = -1;	/* receives no messages, as yet */
			/*
			   task_map[num_tasks].blocking_override=FALSE;
			 */

			strcpy(task_map[num_tasks].ID, ID);

			the_host = find_cpu(host);
			if (the_host < 0) {
				fprintf(stderr,
						  "unknown CPU `%s' line %d\n", host, line_number);
				errors = TRUE;
				goto Errors;
			}

			/* hook up the task to the host processor */
			task_map[num_tasks].host = the_host;

			the_trans = find_trans(trans_ID);
			if (the_trans == -1) {
				fprintf(stderr,
						  "error: unknown transaction ID line %d\n",
						  line_number);
			} else {
				/* if Oi > Ti then an error */
				Ti = trans_map[the_trans].T;
				if (Oi > Ti) {
					fprintf(stderr,
							  "error: Oi(=%d)>Ti(=%d), task %s, line %d\n",
							  Oi, Ti, task_map[num_tasks].ID, line_number);
					errors = TRUE;
				}
				/* need to find transaction period so that task period can
				 * be set; note that task period is equal to ei*Tt
				 */
				task_map[num_tasks].T = Ti * ei;
			}

			/* connect up */
			task_map[num_tasks].trans = the_trans;
			trans_map[the_trans].task_map[trans_map[the_trans].
													num_tasks] = num_tasks;
			trans_map[the_trans].num_tasks++;

			/* keep track of CPU utilisation: more than 100% is bad! */
			utilisation += (double) Ci / (double) (task_map[num_tasks].T);

			/* we've now added the new task in */
			num_tasks++;

		}
		if (!strcmp("blocking", keyword)) {
			int B_factor;
			int the_task;

			sscanf(line, "%s %s %d", keyword, ID, &B_factor);
			the_task = find_task(ID);
			if (the_task < 0) {
				fprintf(stderr,
						  "unknown task `%s' line %d\n", ID, line_number);
				errors = TRUE;
				goto Errors;
			}
			task_map[the_task].blocking_override = TRUE;
			task_map[the_task].B = B_factor;
		}
		if (!strcmp("sem", keyword)) {
			sscanf(line, "%s %s", keyword, ID);
			strcpy(sem_map[num_sems].ID, ID);
			num_sems++;
		}
		if (!strcmp("parameters", keyword)) {
			sscanf(line, "%s %d %d %d %d %d",
					 keyword, &Tclk, &Cclk, &CQL, &CQS, &max_resp);
		}
		if (!strcmp("trans", keyword)) {
			char trans_ID[MAX_STRING];

			sscanf(line, "%s %s %d", keyword, trans_ID, &Ti);
			strcpy(trans_map[num_trans].ID, trans_ID);
			trans_map[num_trans].T = Ti;
			trans_map[num_trans].num_tasks = 0;
			trans_map[num_trans].num_messages = 0;
			num_trans++;
		}
		if (!strcmp("lock", keyword)) {
			int C;
			int the_task;
			int the_sem;
			char sem_ID[MAX_STRING];
			char sugar[MAX_STRING];
			char task_ID[MAX_STRING];

			sscanf(line, "%s %s %s %s %d",
					 keyword, sem_ID, sugar, task_ID, &C);

			if (strcmp(sugar, "by")) {
				fprintf(stderr, "error: expected `by' line %d\n", line_number);
				errors = TRUE;
			}

			the_sem = find_sem(sem_ID);
			if (the_sem < 0) {
				fprintf(stderr,
						  "unknown semaphore `%s' line %d\n",
						  sem_ID, line_number);
				errors = TRUE;
				goto Errors;
			}

			the_task = find_task(task_ID);
			if (the_task < 0) {
				fprintf(stderr,
						  "unknown task `%s' line %d\n", task_ID, line_number);
				errors = TRUE;
				goto Errors;
			}
			lock_map[num_locks].sem = the_sem;
			lock_map[num_locks].task = the_task;
			lock_map[num_locks].C = C;
			num_locks++;
		}
	 Errors:
		;
	}
	/* add one more transaction: create an idle transaction */
	idle_trans = num_trans;
	idle_task = num_tasks;

	strcpy(trans_map[idle_trans].ID, "idletrans");
	strcpy(task_map[idle_task].ID, "idletask");
	trans_map[idle_trans].num_tasks = 1;
	trans_map[idle_trans].task_map[0] = idle_task;
	task_map[idle_task].trans = idle_trans;
	trans_map[idle_trans].T = LARGE_INT;
	task_map[idle_task].priority = num_tasks;
	task_map[idle_task].C = 1;
	task_map[idle_task].T = LARGE_INT;
	task_map[idle_task].e = 1;
	task_map[idle_task].O = 0;
	task_map[idle_task].B = 0;
	task_map[idle_task].J = 0;
	task_map[idle_task].D = LARGE_INT;

	num_trans++;
	num_tasks++;

	utilisation += (double) Cclk / (double) Tclk;

	return errors;
}

int max(a, b)
int a, b;
{
	if (a > b) {
		return a;
	} else {
		return b;
	}
}

int min(a, b)
int a, b;
{
	if (a < b) {
		return a;
	} else {
		return b;
	}
}

timeval simple_tick_overheads(w)
timeval w;
{
	int Lq, Lt;
	int t;

	if (!add_tick_overheads_flag) {
		return 0;
	}

	Lq = 0;
	for (t = 0; t < num_trans; t++) {
		/* add on the queue movements due to transaction t, except for
		 * idle transaction
		 */
		if (t != idle_trans) {
			int x;

			for (x = 0; x < trans_map[t].num_tasks; x++) {
				int j, Jj, Tj;

				j = trans_map[t].task_map[x];
				Jj = task_map[j].J;
				Tj = task_map[j].T;

				Lq += ceildiv(w + Jj, Tj);
			}
		}
	}

	Lt = ceildiv(w, Tclk);

	return Lt * Cclk + min(Lt, Lq) * CQL + max(Lq - Lt, 0) * CQS;
}

/* returns the overheads from clock ticks over a window of width w 
 *
 * assumes that CQS<=CQL
 */
timeval tick_overheads(w)
timeval w;
{
	int Lq, Lt;
	int t;

	if (!add_tick_overheads_flag) {
		return 0;
	}

	Lq = 0;
	for (t = 0; t < num_trans; t++) {
		/* add on the queue movements due to transaction t, except for
		 * idle transaction
		 */
		if (t != idle_trans) {
			int x;

			for (x = 0; x < trans_map[t].num_tasks; x++) {
				int j;

				j = trans_map[t].task_map[x];
				Lq += K(j, t, w);
			}
		}
	}

	Lt = ceildiv(w, Tclk);

	return Lt * Cclk + min(Lt, Lq) * CQL + max(Lq - Lt, 0) * CQS;
}

/* tries to find CPU indicated by alphanumeric ID; returns -1
 * if cannot find the CPU
 */
int find_cpu(host_ID)
char *host_ID;
{
	int i;

	for (i = 0; i < num_cpus; i++) {
		if (!strcmp(cpu_map[i].ID, host_ID)) {
			return i;
		}
	}

	return -1;
}

/* tries to find network indicated by alphanumeric ID; returns -1
 * if cannot find the network
 */
int find_net(net_ID)
char *net_ID;
{
	int i;

	for (i = 0; i < num_nets; i++) {
		if (!strcmp(net_map[i].ID, net_ID)) {
			return i;
		}
	}

	return -1;
}

/* tries to find transaction indicated by alphanumeric ID; returns -1
 * if cannot find the transaction
 */
int find_trans(trans_ID)
char *trans_ID;
{
	int i;

	for (i = 0; i < num_trans; i++) {
		if (!strcmp(trans_map[i].ID, trans_ID)) {
			return i;
		}
	}

	return -1;
}

/* tries to find semaphore indicated by alphanumeric ID; returns -1
 * if cannot find the semaphore
 */
int find_sem(sem_ID)
char *sem_ID;
{
	int i;

	for (i = 0; i < num_sems; i++) {
		if (!strcmp(sem_map[i].ID, sem_ID)) {
			return i;
		}
	}

	return -1;
}

/* tries to find task indicated by alphanumeric ID; returns -1
 * if cannot find the task
 */
int find_task(task_ID)
char *task_ID;
{
	int i;

	for (i = 0; i < num_tasks; i++) {
		if (!strcmp(task_map[i].ID, task_ID)) {
			return i;
		}
	}

	return -1;
}

/* return the blocking factor of task i:
 *
 * "the longest critical section of a lower priority task accessing a
 * semaphore with ceiling >= priority of i"
 */
int blocking_factor(i)
int i;
{
	int max_cs = 0;				  /* largest critical section */
	int j;

	if (task_map[i].blocking_override) {
		return task_map[i].B;
	}

	for (j = 0; j < num_locks; j++) {	/* look down the lock map */
		int the_task;				  /* task doing the locking */
		int the_sem;				  /* the semaphore it locks */
		int the_ceiling;			  /* the ceiling of the semaphore */
		int the_cs;					  /* the computation time the_task holds the lock for */
		int priority_i = task_map[i].priority;	/* priority of task i */

		the_task = lock_map[j].task;
		the_sem = lock_map[j].sem;
		the_ceiling = sem_map[the_sem].ceiling;
		the_cs = lock_map[j].C;
		if (the_ceiling <= priority_i
			 && task_map[the_task].priority > priority_i) {
			max_cs = max(max_cs, the_cs);
		}
	}
	return max_cs;
}

/* find the ceiling of semaphore s */
void set_ceiling(s)
int s;
{
	int i;

	sem_map[s].ceiling = num_tasks - 1;	/* set ceiling to lowest priority */
	for (i = 0; i < num_locks; i++) {
		if (lock_map[i].sem == s) {
			sem_map[s].ceiling =
				 min(sem_map[s].ceiling, task_map[lock_map[i].task].priority);
		}
	}
}

void set_all_Bi()
{
	int i;

	/* compute semaphore ceilings */
	for (i = 0; i < num_sems; i++) {
		set_ceiling(i);
	}
	/* now compute blocking factors */
	for (i = 0; i < num_tasks; i++) {
		task_map[i].B = blocking_factor(i);
	}
}

void print_details(stream)
FILE *stream;
{
	int i, t;

	fprintf(stream, "utilisation=%f%%\n", utilisation * 100.0);
	fprintf(stream, "Tclk=%d, Cclk=%d, CQL=%d, CQS=%d\n", Tclk,
			  Cclk, CQL, CQS);
	fprintf(stream,
			  "tick overheads in a window of size %d are %d\n",
			  tick_window, tick_overheads(tick_window));

	if (strchr(output_reqd, 't')) {
		int j;

		for (j = 0; j < num_tasks; j++) {
			for (i = 0; i < num_tasks; i++) {
				if (task_map[i].priority == j) {
					if (task_map[i].r < task_map[i].maxfoundri) {
						fprintf(stderr,
								  "failure: sim r(%d) != analysis r(%d)\n",
								  task_map[i].maxfoundri, task_map[i].r);
						fprintf(stderr,
								  "(failed for task %s)\n", task_map[i].ID);
					}
					print_task_attribute(stdout, i);
					print_task_attribute(word_file, i);
				}
			}
		}

		for (j = 0; j < num_messages; j++) {
			print_message_attribute(stdout, j);
		}

		for (t = 0; t < num_trans; t++) {
			int x;

			fprintf(output_file, "%s, T:%d\n",
					  trans_map[t].ID, trans_map[t].T);
			for (x = 0; x < trans_map[t].num_tasks; x++) {
				fprintf(output_file, "\t%s\n",
						  task_map[trans_map[t].task_map[x]].ID);
				if (word_file) {
					fprintf(word_file, ",%s\n",
							  task_map[trans_map[t].task_map[x]].ID);
				}
			}
			if (word_file) {
				fprintf(word_file, "%s,%d\n", trans_map[t].ID, trans_map[t].T);
			}
		}
	}

	if (strchr(output_reqd, 'o')) {
		for (i = 0; i < num_sems; i++) {
			fprintf(stream,
					  "%s ceiling %d\n", sem_map[i].ID, sem_map[i].ceiling);
		}
		for (i = 0; i < num_locks; i++) {
			fprintf(stream,
					  "%s locked by %s for time %d\n",
					  sem_map[lock_map[i].sem].ID,
					  task_map[lock_map[i].task].ID, lock_map[i].C);
			if (lock_map[i].C > task_map[lock_map[i].task].C) {
				fprintf(stderr,
						  "consistency: %s locks %s for longer than WCET!\n",
						  task_map[lock_map[i].task].ID,
						  sem_map[lock_map[i].sem].ID);
			}
		}
	}

	if (strchr(output_reqd, 'c')) {
		for (i = 0; i < num_precs; i++) {
			if (prec_map[i].result == FALSE) {
				printf
					 ("precedence constraint between %s and %s not met!\n",
					  task_map[prec_map[i].a].ID, task_map[prec_map[i].b].ID);
			} else {
				printf
					 ("precedence constraint between %s and %s met\n",
					  task_map[prec_map[i].a].ID, task_map[prec_map[i].b].ID);
			}
		}
		for (i = 0; i < num_excls; i++) {
			if (excl_map[i].result == FALSE) {
				printf
					 ("exclusion constraint between %s and %s not met!\n",
					  task_map[excl_map[i].a].ID, task_map[excl_map[i].b].ID);
			} else {
				printf
					 ("exclusion constraint between %s and %s met\n",
					  task_map[excl_map[i].a].ID, task_map[excl_map[i].b].ID);
			}
		}
	}
}

/* simulate the system for t ticks */
void simulate(t)
int t;
{
	sim_init();
	do {
		sim_stretch();
	}
	while (clock_tick < t);
}

int call_logs;

/* initialise tasks: all not runnable, with next run set appropriately
 */
void sim_init()
{
	int i;

	clock_tick = 0;
	for (i = 0; i < num_tasks; i++) {
		if (i == idle_task) {
			/* idle is always runnable */
			task_map[i].runnable = TRUE;
		} else {
			task_map[i].runnable = FALSE;
		}
		task_map[i].runat = task_map[i].O;
		task_map[i].maxfoundri = 0;
		task_map[i].remaining = task_map[i].C;
	}

	call_logs = 0;
}

/* simulate a stretch of time; clock_tick always gives the clock */
void sim_stretch()
{
	int i;
	int ti_cpu;
	int highest_runnable[MAX_CPUS];
	int earlytask;
	int earlyevent = LARGE_INT;

	call_logs++;
	if (call_logs > 100) {
		call_logs = 0;
		/* printf("%d.. ", clock_tick); */
	}

	for (i = 0; i < num_cpus; i++) {
		highest_runnable[i] = idle_task;
	}
	for (i = 0; i < num_tasks; i++) {
		ti_cpu = task_map[i].host;
		/* check to see if a task is runnable */
		if (task_map[i].runnable == FALSE && clock_tick >= task_map[i].runat) {
			task_map[i].runnable = TRUE;
		}

		/* find the highest priority runnable task */
		if (task_map[i].priority <
			 task_map[highest_runnable[ti_cpu]].priority
			 && task_map[i].runnable) {
			/* we have selected a potentially runnable task */
			highest_runnable[ti_cpu] = i;
		}
	}
/*      for(i=0; i<num_tasks; i++) */
/*      { */
	/* find earliest event where the about-to-be-run task could be
	 * pre-empted
	 */
/*  	if(task_map[i].runnable==FALSE && task_map[i].runat<earlyevent */
/*  	   && task_map[i].priority<task_map[highest_runnable[task_map[i].host]].priority) */
/*  	{ */
/*  	    earlyevent=task_map[i].runat; */
/*  	    earlytask=i; */
/*  	} */
/*      } */

	clock_tick++;
	for (ti_cpu = 0; ti_cpu < num_cpus; ti_cpu++) {
		if (highest_runnable[ti_cpu] == idle_task) {
			/* don't want to do anything */
		} else {
			int duration;
			/* want to run the task until either earlyevent, or until out of
			 * time (in which case, placed back on the queue)
			 */
			i = highest_runnable[ti_cpu];
/*  	    duration=min(task_map[i].remaining, */
/*  			 earlyevent-clock_tick); */

/*  	    task_map[i].remaining-=duration; */
/*  	    clock_tick+=duration; */

			task_map[i].remaining--;
			if (task_map[i].remaining == 0) {
				/* put back on the delay queue */
				task_map[i].runnable = FALSE;

				/* record the worst response time */
				task_map[i].maxfoundri =
					 max(task_map[i].maxfoundri,
						  clock_tick - task_map[i].runat);

				/* replenish CPU budget */
				task_map[i].remaining = task_map[i].C;

				/* place next runnable time back on queue */
				task_map[i].runat = task_map[i].runat + task_map[i].T;
				if (task_map[i].T !=
					 task_map[i].e * trans_map[task_map[i].trans].T) {
					fprintf(stderr, "simulation assertion failure!\n");
					exit(1);
				}
			}
		}
	}
}

/* check precedence constraint is met by two tasks, a and b
 * return TRUE if constraint met, FALSE otherwise
 */
int check_prec(a, b)
int a, b;
{
	timeval Oa, Ob, Ja;
	timeval ra;

	Oa = task_map[a].O;
	Ob = task_map[b].O;
	Ja = task_map[a].J;
	ra = task_map[a].r;

	/* assertion check: both tasks must be members of the same
	 * transaction
	 */
	if (task_map[a].trans != task_map[b].trans) {
		fprintf(stderr, "precedence assertion failure!\n");
		printf("task a is %s, task b is %s\n", task_map[a].ID,
				 task_map[b].ID);
		exit(1);
	}

	if (task_map[a].priority < task_map[b].priority) {
		/* use the first condition [equation 30] */
		if (Oa + Ja <= Ob) {
			return TRUE;
		} else {
			return FALSE;
		}
	} else {
		/* use the second condition [equation 31] */
		if (Oa + ra < Ob) {
			return TRUE;
		} else {
			return FALSE;
		}
	}
}

/* check the exclusion constraint between a and b holds */
int check_excl(a, b)
int a, b;
{
	timeval Tt;
	timeval Oa, Ja, Ob, rb;

	/* we require a to be of higher priority; if not then swap a and b */
	if (task_map[b].priority < task_map[a].priority) {
		/* swap */
		int tmp;
		tmp = a;
		a = b;
		b = tmp;
	}

	Oa = task_map[a].O;
	Ob = task_map[b].O;
	Ja = task_map[a].J;
	rb = task_map[b].r;
	Tt = trans_map[task_map[a].trans].T;

	/* assertion check: both tasks must be members of the same
	 * transaction
	 */
	if (task_map[a].trans != task_map[b].trans) {
		fprintf(stderr, "exclusion assertion failure!\n");
		exit(1);
	}

	if (Oa + Ja <= Ob && Ob + rb < Oa + Tt) {
		return TRUE;
	} else {
		return FALSE;
	}
}


/* prints to the output stream the attributes of task i */
void print_task_attribute(stream, i)
FILE *stream;
int i;
{
	int j;

	static char s[] = "x:";

	if (stream == NULL) {
		return;
	}
	if (stream == word_file) {
		fprintf(stream, "%s,", task_map[i].ID);
	} else {
		fprintf(stream, "%s(%s,%s),", task_map[i].ID,
				  trans_map[task_map[i].trans].ID,
				  cpu_map[task_map[i].host].ID);
		fprintf(stream, "end:%d,", task_map[i].O + task_map[i].r);
	}

	for (j = 0; j < strlen(output_reqd); j++) {
		s[0] = output_reqd[j];
		switch (output_reqd[j]) {
		case 'r':
			fprintf(stream, "%s%d,", s, task_map[i].r);
			break;
		case 'R':
			fprintf(stream, "%s%d,", s, task_map[i].dmsr);
			break;
		case 'D':
			if (task_map[i].D < 0) {
				fprintf(stream, "X,");
			} else {
				fprintf(stream, "%s%d,", s, task_map[i].D);
			}
			break;
		case 'T':
			fprintf(stream, "%s%d,", s, task_map[i].T);
			break;
		case 'e':
			fprintf(stream, "%s%d,", s, task_map[i].e);
			break;
		case 'B':
			fprintf(stream, "%s%d,", s, task_map[i].B);
			break;
		case 'C':
			fprintf(stream, "%s%d,", s, task_map[i].C);
			break;
		case 'O':
			fprintf(stream, "%s%d,", s, task_map[i].O);
			break;
		case 'J':
			fprintf(stream, "%s%d,", s, task_map[i].J);
			break;
		}
	}
	fprintf(stream, "\n");
}

void print_message_attribute(stream, i)
FILE *stream;
int i;
{
	int j;
	static char s[] = "x:";

	if (stream == NULL) {
		return;
	}
	if (stream == word_file) {
		fprintf(stream, "%s,", task_map[i].ID);
	} else {
		fprintf(stream, "%s(%s,%s),", message_map[i].ID,
				  trans_map[message_map[i].trans].ID,
				  net_map[message_map[i].net].ID);
		fprintf(stream, "end:%d,", message_map[i].O + message_map[i].r);
	}
	for (j = 0; j < strlen(output_reqd); j++) {
		s[0] = output_reqd[j];
		switch (output_reqd[j]) {
		case 'r':
			fprintf(stream, "%s%d,", s, message_map[i].r);
			break;
		case 'R':
			fprintf(stream, "%s%d,", s, message_map[i].r);
			break;
		case 'D':
			if (message_map[i].D < 0) {
				fprintf(stream, "X,");
			} else {
				fprintf(stream, "%s%d,", s, message_map[i].D);
			}
			break;
		case 'T':
			fprintf(stream, "%s%d,", s, message_map[i].T);
			break;
		case 'e':
			fprintf(stream, "%s%d,", s, message_map[i].e);
			break;
		case 'B':
			fprintf(stream, "%s%d,", s, message_map[i].B);
			break;
		case 'C':
			fprintf(stream, "%s%d,", s, message_map[i].C);
			break;
		case 'O':
			fprintf(stream, "%s%d,", s, message_map[i].O);
			break;
		case 'J':
			fprintf(stream, "%s%d,", s, message_map[i].J);
			break;
		}
	}
	fprintf(stream, "\n");
}
